'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

var vue = require('vue');

class ValidationConfig {
    constructor() {
        Object.defineProperty(this, "defaultValidationBehavior", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: void 0
        });
        Object.defineProperty(this, "validationBehavior", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: void 0
        });
        const aggressive = () => true;
        const lazy = ({ touched }) => touched;
        const lazier = ({ force, touched, submit, hasError }) => force || submit || (touched && hasError);
        const submit = ({ submit, hasError }) => submit || hasError;
        this.defaultValidationBehavior = 'lazier';
        this.validationBehavior = new Map([
            ['aggressive', aggressive],
            ['lazy', lazy],
            ['lazier', lazier],
            ['submit', submit]
        ]);
    }
    getDefaultValidationBehavior() {
        return this.validationBehavior.get(this.defaultValidationBehavior);
    }
}
const VALIDATION_CONFIG = new ValidationConfig();

class LinkedListNode {
    constructor(value) {
        Object.defineProperty(this, "value", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: void 0
        });
        Object.defineProperty(this, "next", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: null
        });
        Object.defineProperty(this, "prev", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: null
        });
        this.value = value;
    }
}
class LinkedList {
    constructor() {
        Object.defineProperty(this, "first", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: null
        });
        Object.defineProperty(this, "last", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: null
        });
        Object.defineProperty(this, "count", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: 0
        });
    }
    addFirst(value) {
        const node = new LinkedListNode(value);
        if (this.count === 0) {
            this.first = node;
            this.last = node;
        }
        else {
            node.next = this.first;
            this.first.prev = node;
            this.first = node;
        }
        this.count++;
        return node;
    }
    addLast(value) {
        const node = new LinkedListNode(value);
        if (this.count === 0) {
            this.first = node;
            this.last = node;
        }
        else {
            node.prev = this.last;
            this.last.next = node;
            this.last = node;
        }
        this.count++;
        return node;
    }
    remove(node) {
        if (this.count === 0) {
            return;
        }
        if (node === this.first) {
            this.removeFirst();
        }
        else if (node === this.last) {
            this.removeLast();
        }
        else {
            node.prev.next = node.next;
            node.next.prev = node.prev;
            node.next = null;
            node.prev = null;
            this.count--;
        }
    }
    removeFirst() {
        if (this.count === 0) {
            return;
        }
        if (this.count === 1) {
            this.first = null;
            this.last = null;
            this.count--;
        }
        else {
            this.first = this.first.next;
            this.first.prev.next = null;
            this.first.prev = null;
            this.count--;
        }
    }
    removeLast() {
        if (this.count === 0) {
            return;
        }
        if (this.count === 1) {
            this.first = null;
            this.last = null;
            this.count--;
        }
        else {
            this.last = this.last.prev;
            this.last.next.prev = null;
            this.last.next = null;
            this.count--;
        }
    }
    *nodesForwards() {
        let node = this.first;
        for (; node !== null; node = node.next) {
            yield node;
        }
    }
    *nodesBackwards() {
        let node = this.last;
        for (; node !== null; node = node.prev) {
            yield node;
        }
    }
}

const isDefined = (x) => x !== null && typeof x !== 'undefined';
const isRecord = (x) => typeof x === 'object' && x !== null && !Array.isArray(x);
const isArray = (x) => Array.isArray(x);
const isObject = (x) => typeof x === 'object' && x !== null;

function* deepIterator(obj, predicate = () => false) {
    const stack = new LinkedList();
    stack.addLast({ current: obj, parent: null, parentKey: '', path: [] });
    while (stack.count > 0) {
        const { current, parent, parentKey, path } = stack.last.value;
        stack.removeLast();
        let pushedItemsOnStack = false;
        if (isObject(current) && !vue.isRef(current) && !predicate(current)) {
            const entries = Object.entries(current);
            pushedItemsOnStack = entries.length > 0;
            for (let i = entries.length - 1; i >= 0; i--) {
                const [key, value] = entries[i];
                stack.addLast({
                    current: value,
                    parent: current,
                    parentKey: key,
                    path: [...path, key]
                });
            }
        }
        if (isObject(parent)) {
            yield {
                key: parentKey,
                value: parent[parentKey],
                parent,
                path,
                isLeaf: !pushedItemsOnStack
            };
        }
    }
}

function set(obj, keys, value) {
    if (keys.length === 0) {
        return;
    }
    let o = obj;
    for (let i = 0; i < keys.length - 1; i++) {
        const key = keys[i];
        const nextKey = keys[i + 1];
        const value = o[key];
        if (value === undefined) {
            if (Number.isNaN(+nextKey)) {
                o[key] = {};
            }
            else {
                o[key] = [];
            }
        }
        o = o[key];
    }
    o[keys[keys.length - 1]] = value;
}

function deepCopy(toClone) {
    if (isObject(toClone)) {
        const copy = isArray(toClone) ? [] : {};
        for (const { value, path, isLeaf } of deepIterator(toClone)) {
            if (isLeaf) {
                set(copy, path, value);
            }
        }
        return copy;
    }
    return toClone;
}

const trySet = (map) => ({ success, failure }) => (key, value) => {
    const _value = map.get(key);
    if (_value) {
        failure === null || failure === void 0 ? void 0 : failure(_value);
    }
    else {
        map.set(key, value);
        success === null || success === void 0 ? void 0 : success(value);
    }
};
const tryGet = (map) => ({ success, failure }) => (key) => {
    const value = map.get(key);
    if (value) {
        success(value);
    }
    else {
        failure === null || failure === void 0 ? void 0 : failure();
    }
};

function path(path, obj) {
    let value = obj[path[0]];
    for (let i = 0; i < path.length; i++) {
        const key = path[i];
        if (value === null || value === undefined) {
            return undefined;
        }
        if (i > 0) {
            value = value[key];
        }
    }
    return value;
}

let id = 1;
function uid() {
    return id++;
}

function debounce(target, { wait, shouldInvoke }) {
    let timerId = null;
    let shouldInvokeResult = true;
    function debounced(...args) {
        if (shouldInvoke) {
            shouldInvokeResult = shouldInvoke.apply(this, args);
        }
        const effect = (shouldInvoke) => () => {
            timerId = null;
            if (shouldInvoke === true) {
                target.apply(this, args);
            }
        };
        if (shouldInvokeResult === true) {
            clearTimeout(timerId);
            timerId = setTimeout(effect(shouldInvokeResult), wait);
        }
        else {
            timerId = null;
        }
    }
    return debounced;
}

class PromiseCancel {
    constructor() {
        Object.defineProperty(this, "promise", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: void 0
        });
        Object.defineProperty(this, "resolve", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: void 0
        });
        Object.defineProperty(this, "reject", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: void 0
        });
        this.assign();
    }
    cancelResolve(value) {
        this.resolve(value);
        this.assign();
    }
    cancelReject(reason) {
        this.reject(reason);
        this.assign();
    }
    race(...promises) {
        return Promise.race([this.promise, ...promises]);
    }
    assign() {
        this.promise = new Promise((resolve, reject) => {
            this.resolve = resolve;
            this.reject = reject;
        });
    }
}

const isField = (x) => isRecord(x) ? '$value' in x : false;
const isTransformedField = (x) => isRecord(x) ? '$uid' in x && '$value' in x : false;
function mapFieldRules(fieldRules) {
    const defaultValidationBehavior = VALIDATION_CONFIG.getDefaultValidationBehavior();
    return fieldRules.map(fieldRule => {
        if (typeof fieldRule === 'function') {
            return {
                validationBehavior: defaultValidationBehavior,
                rule: fieldRule
            };
        }
        if (Array.isArray(fieldRule)) {
            const [first, second, third] = fieldRule;
            if (typeof second === 'number') {
                return {
                    validationBehavior: defaultValidationBehavior,
                    rule: first,
                    debounce: second
                };
            }
            if (typeof first === 'function') {
                return {
                    validationBehavior: first,
                    rule: second,
                    debounce: third
                };
            }
            const validationBehavior = VALIDATION_CONFIG.validationBehavior.get(first);
            if (validationBehavior !== undefined) {
                return { validationBehavior, rule: second, debounce: third };
            }
            else {
                console.warn(`[useValidation] Validation behavior with name '${first}' does not exist. Valid values are`, VALIDATION_CONFIG.validationBehavior.keys());
            }
        }
        else {
            return {
                validationBehavior: defaultValidationBehavior,
                rule: fieldRule
            };
        }
        throw Error('[useValidation] Invalid rule provided');
    });
}
function registerField(form, name, field) {
    const { $value, $rules, ...fieldExtraProperties } = field;
    const rules = $rules ? mapFieldRules($rules) : [];
    const uid$1 = uid();
    const formField = form.registerField(uid$1, name, $value, rules);
    return {
        ...fieldExtraProperties,
        $uid: uid$1,
        $value: formField.modelValue,
        $errors: formField.errors,
        $hasError: formField.hasError,
        $rawErrors: formField.rawErrors,
        $validating: formField.validating,
        $dirty: formField.dirty,
        $touched: formField.touched,
        $validate({ setTouched, force } = {}) {
            setTouched !== null && setTouched !== void 0 ? setTouched : (setTouched = true);
            force !== null && force !== void 0 ? force : (force = true);
            if (setTouched) {
                formField.touched.value = true;
            }
            form.validate(uid$1, force);
        }
    };
}
function transformFormData(form, formData) {
    for (const { key, value, parent } of deepIterator(formData)) {
        if (isField(value)) {
            const transformedField = registerField(form, key, value);
            parent[key] = transformedField;
        }
    }
}
function getResultFormData(transformedFormData, predicate = () => true) {
    const result = {};
    for (const { key, value, path, isLeaf } of deepIterator(transformedFormData, isTransformedField)) {
        if (isLeaf) {
            const unpackedValue = isTransformedField(value)
                ? value.$value
                : vue.unref(value);
            if (predicate({ key, value: unpackedValue, path }) === true) {
                // Value is reactive -> value is an object or array
                // Make sure to do a deep clone to loose the reactive reference
                if (vue.isReactive(unpackedValue)) {
                    set(result, path, deepCopy(unpackedValue));
                }
                else {
                    set(result, path, unpackedValue);
                }
            }
        }
    }
    return result;
}
function disposeForm(form, deletedFormData) {
    for (const { value } of deepIterator({ box: deletedFormData }, isTransformedField)) {
        if (isTransformedField(value)) {
            form.dispose(value.$uid);
        }
    }
}
function resetFields(form, data, transformedFormData) {
    Object.entries(data).forEach(([key, value]) => {
        const transformedValue = transformedFormData[key];
        if (isTransformedField(transformedValue)) {
            const field = form.getField(transformedValue.$uid);
            field.reset(value);
            return;
        }
        if (isObject(value)) {
            resetFields(form, value, transformedFormData[key]);
        }
    });
}

const isSimpleRule = (rule) => typeof rule === 'function';

class FormField {
    constructor(form, uid, name, modelValue, ruleInfos) {
        Object.defineProperty(this, "_rules", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: void 0
        });
        Object.defineProperty(this, "_validationBehaviors", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: void 0
        });
        Object.defineProperty(this, "_buffers", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: void 0
        });
        Object.defineProperty(this, "_watchStopHandle", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: void 0
        });
        Object.defineProperty(this, "_form", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: void 0
        });
        Object.defineProperty(this, "uid", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: void 0
        });
        Object.defineProperty(this, "rulesValidating", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: vue.ref(0)
        });
        Object.defineProperty(this, "initialModelValue", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: void 0
        });
        Object.defineProperty(this, "name", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: void 0
        });
        Object.defineProperty(this, "touched", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: vue.ref(false)
        });
        Object.defineProperty(this, "dirty", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: vue.ref(false)
        });
        Object.defineProperty(this, "modelValue", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: void 0
        });
        Object.defineProperty(this, "rawErrors", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: void 0
        });
        Object.defineProperty(this, "errors", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: vue.computed(() => this.rawErrors.filter(isDefined))
        });
        Object.defineProperty(this, "validating", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: vue.computed(() => this.rulesValidating.value > 0)
        });
        Object.defineProperty(this, "hasError", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: vue.computed(() => this.errors.value.length > 0)
        });
        this._rules = ruleInfos.map(({ rule }) => isSimpleRule(rule) ? rule : rule.rule);
        this._validationBehaviors = ruleInfos.map(info => info.validationBehavior);
        this._buffers = ruleInfos.map(() => new LinkedList());
        this._form = form;
        this.rawErrors = vue.reactive(ruleInfos.map(() => null));
        this.name = name;
        this.uid = uid;
        if (vue.isRef(modelValue) || vue.isReactive(modelValue)) {
            this.modelValue = modelValue;
            this.initialModelValue = deepCopy(vue.unref(modelValue));
        }
        else if (isRecord(modelValue)) {
            this.modelValue = vue.reactive(modelValue);
            this.initialModelValue = deepCopy(this.modelValue);
        }
        else {
            this.modelValue = vue.ref(modelValue);
            this.initialModelValue = deepCopy(vue.unref(modelValue));
        }
        this._watchStopHandle = vue.watch(this.modelValue, () => {
            this.dirty.value = true;
            this._form.validate(this.uid);
        }, { deep: true });
    }
    async validate(ruleNumber, modelValues, noThrow) {
        const rule = this._rules[ruleNumber];
        if (!rule) {
            return;
        }
        const buffer = this._buffers[ruleNumber];
        let error;
        const ruleResult = rule(...modelValues.map(vue.unref));
        const shouldSetError = buffer.addLast(true);
        if (shouldSetError.prev) {
            shouldSetError.prev.value = false;
            this.rulesValidating.value--;
            this._form.rulesValidating.value--;
        }
        if (typeof (ruleResult === null || ruleResult === void 0 ? void 0 : ruleResult.then) === 'function') {
            this.rulesValidating.value++;
            this._form.rulesValidating.value++;
            try {
                error = await ruleResult;
            }
            catch (err) {
                error = err;
            }
            buffer.remove(shouldSetError);
            if (shouldSetError.value) {
                this.rulesValidating.value--;
                this._form.rulesValidating.value--;
                this._setError(ruleNumber, error, noThrow);
            }
        }
        else {
            buffer.removeLast();
            error = ruleResult;
            this._setError(ruleNumber, error, noThrow);
        }
    }
    reset(resetValue = this.initialModelValue) {
        this._watchStopHandle();
        this.touched.value = false;
        this.dirty.value = false;
        if (vue.isRef(this.modelValue)) {
            if (isArray(this.modelValue.value)) {
                this.modelValue.value = deepCopy(resetValue);
            }
            else {
                this.modelValue.value = resetValue;
            }
        }
        else {
            const copy = deepCopy(resetValue);
            Object.assign(this.modelValue, copy);
        }
        this.rulesValidating.value = 0;
        this._form.rulesValidating.value = 0;
        for (const buffer of this._buffers) {
            for (const shouldSetError of buffer.nodesForwards()) {
                shouldSetError.value = false;
            }
        }
        Object.assign(this.rawErrors, this.rawErrors.map(() => null));
        this._watchStopHandle = vue.watch(this.modelValue, () => {
            this.dirty.value = true;
            this._form.validate(this.uid);
        }, { deep: true });
    }
    dispose() {
        this.errors.effect.stop();
        this.validating.effect.stop();
        this.hasError.effect.stop();
        this._watchStopHandle();
    }
    shouldValidate(ruleNumber, force, submit) {
        return this._validationBehaviors[ruleNumber]({
            hasError: this.rawErrors[ruleNumber] !== null,
            touched: this.touched.value,
            dirty: this.dirty.value,
            force,
            submit,
            value: vue.unref(this.modelValue)
        });
    }
    _setError(ruleNumber, error, noThrow) {
        if (typeof error === 'string') {
            this.rawErrors[ruleNumber] = error;
            if (!noThrow) {
                throw error;
            }
        }
        else {
            this.rawErrors[ruleNumber] = null;
        }
    }
}

class ValidationError extends Error {
    constructor() {
        super('One or more validation errors occurred.');
    }
}

class Form {
    constructor() {
        Object.defineProperty(this, "_simpleValidators", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: new Map()
        });
        Object.defineProperty(this, "_keyedValidators", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: new Map()
        });
        Object.defineProperty(this, "_reactiveFieldMap", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: vue.shallowReactive(new Map())
        });
        Object.defineProperty(this, "tryGetSimpleValidators", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: tryGet(this._simpleValidators)
        });
        Object.defineProperty(this, "trySetKeyedValidators", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: trySet(this._keyedValidators)
        });
        Object.defineProperty(this, "tryGetKeyedValidators", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: tryGet(this._keyedValidators)
        });
        Object.defineProperty(this, "rulesValidating", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: vue.ref(0)
        });
        Object.defineProperty(this, "submitting", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: vue.ref(false)
        });
        Object.defineProperty(this, "validating", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: vue.computed(() => this.rulesValidating.value > 0)
        });
        Object.defineProperty(this, "hasError", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: vue.computed(() => this.errors.value.length > 0)
        });
        Object.defineProperty(this, "errors", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: vue.computed(() => {
                const errors = [];
                for (const field of this._reactiveFieldMap.values()) {
                    errors.push(...field.errors.value);
                }
                return errors;
            })
        });
    }
    registerField(uid, name, modelValue, ruleInfos) {
        const field = new FormField(this, uid, name, modelValue, ruleInfos);
        const simpleValidators = {
            validators: [],
            validatorsNotDebounced: [],
            meta: {
                field,
                keys: [],
                rollbacks: []
            }
        };
        ruleInfos.forEach(({ rule, debounce: debounce$1 }, ruleNumber) => {
            const validatorNotDebounced = (modelValues, force, submit) => {
                if (field.shouldValidate(ruleNumber, force, submit) === true) {
                    return field.validate(ruleNumber, modelValues, false);
                }
            };
            let incrementedTimes = 0;
            const validator = debounce$1
                ? debounce(modelValues => {
                    field.validate(ruleNumber, modelValues, true);
                    field.rulesValidating.value -= incrementedTimes;
                    incrementedTimes = 0;
                }, {
                    wait: debounce$1,
                    shouldInvoke(modelValues, force, submit) {
                        if (field.shouldValidate(ruleNumber, force, submit) === true) {
                            field.rulesValidating.value++;
                            incrementedTimes++;
                            return true;
                        }
                    }
                })
                : (modelValues, force, submit) => {
                    if (field.shouldValidate(ruleNumber, force, submit) === true) {
                        field.validate(ruleNumber, modelValues, true);
                    }
                };
            if (isSimpleRule(rule)) {
                simpleValidators.validators.push(validator);
                simpleValidators.validatorsNotDebounced.push(validatorNotDebounced);
            }
            else {
                const { key } = rule;
                const keyedValidator = {
                    validator,
                    validatorNotDebounced,
                    meta: {
                        field
                    }
                };
                const rollback = () => {
                    this.tryGetKeyedValidators({
                        success: keyedValidators => {
                            keyedValidators.delete(keyedValidator);
                            if (keyedValidators.size === 0) {
                                this._keyedValidators.delete(key);
                            }
                        }
                    })(key);
                };
                simpleValidators.meta.keys.push(key);
                simpleValidators.meta.rollbacks.push(rollback);
                this.trySetKeyedValidators({
                    failure: keyedValidators => keyedValidators.add(keyedValidator)
                })(key, new Set([keyedValidator]));
            }
        });
        this._simpleValidators.set(uid, simpleValidators);
        this._reactiveFieldMap.set(uid, field);
        return field;
    }
    validate(uid, force = false) {
        const simpleValidators = this._simpleValidators.get(uid);
        const { validators, meta } = simpleValidators;
        for (const validator of validators) {
            validator([meta.field.modelValue], force, false);
        }
        this._invokeValidatorsForKeys(meta.keys, force, false);
    }
    async validateAll(names) {
        const settledResults = await Promise.allSettled(this._collectValidatorResultsForNames(names));
        for (const result of settledResults) {
            if (result.status === 'rejected') {
                throw new ValidationError();
            }
        }
    }
    dispose(uid) {
        this.tryGetSimpleValidators({
            success: ({ meta }) => {
                meta.field.dispose();
                meta.rollbacks.forEach(r => r());
            }
        })(uid);
        this._simpleValidators.delete(uid);
        this._reactiveFieldMap.delete(uid);
    }
    resetFields() {
        for (const { meta } of this._simpleValidators.values()) {
            meta.field.reset();
        }
    }
    getField(uid) {
        const simpleValidators = this._simpleValidators.get(uid);
        if (simpleValidators) {
            return simpleValidators.meta.field;
        }
    }
    _invokeValidatorsForKeys(keys, force, submit) {
        for (const key of keys) {
            const keyedValidators = this._keyedValidators.get(key);
            if (this._isEveryFieldTouched(keyedValidators)) {
                const values = [...keyedValidators.values()];
                const modelValues = values.map(({ meta }) => meta.field.modelValue);
                for (const { validator } of values) {
                    validator(modelValues, force, submit);
                }
            }
        }
    }
    *_collectValidatorResultsForKeys(keys, force, submit) {
        for (const key of keys) {
            const keyedValidators = this._keyedValidators.get(key);
            const values = [...keyedValidators.values()];
            const modelValues = values.map(({ meta }) => meta.field.modelValue);
            for (const { validatorNotDebounced } of values) {
                yield validatorNotDebounced(modelValues, force, submit);
            }
        }
    }
    *_collectValidatorResultsForNames(names) {
        if (names === undefined) {
            for (const { validatorsNotDebounced, meta } of this._simpleValidators.values()) {
                meta.field.touched.value = true;
                for (const validator of validatorsNotDebounced) {
                    yield validator([meta.field.modelValue], false, true);
                }
            }
            yield* this._collectValidatorResultsForKeys(this._keyedValidators.keys(), false, true);
        }
        else if (names.length > 0) {
            const uniqueNames = new Set(names);
            for (const { validatorsNotDebounced, meta } of this._simpleValidators.values()) {
                if (uniqueNames.has(meta.field.name)) {
                    meta.field.touched.value = true;
                    for (const validator of validatorsNotDebounced) {
                        yield validator([meta.field.modelValue], false, true);
                    }
                    yield* this._collectValidatorResultsForKeys(this._keyedValidators.keys(), false, true);
                }
            }
        }
    }
    _isEveryFieldTouched(keyedValidators) {
        for (const { meta } of keyedValidators) {
            if (!meta.field.touched.value) {
                return false;
            }
        }
        return true;
    }
}

/**
 *
 * Vue composition function for form validation.
 *
 * @remarks
 * For type inference inside of `useValidation` you need to define the structure of your
 * `formData` upfront and pass it as the generic parameter `FormData`.
 *
 * @param formData - The structure of your `formData`.
 *
 * @example
 * ```
 * type FormData = {
 *   name: Field<string>,
 *   password: Field<string>
 * }
 *
 * const { form } = useValidation<FormData>({})
 * ```
 */
function useValidation(formData) {
    const form = new Form();
    const promiseCancel = new PromiseCancel();
    transformFormData(form, formData);
    const transformedFormData = vue.reactive(formData);
    return {
        form: transformedFormData,
        submitting: form.submitting,
        validating: form.validating,
        hasError: form.hasError,
        errors: form.errors,
        async validateFields({ names, predicate } = {}) {
            form.submitting.value = true;
            const resultFormData = getResultFormData(transformedFormData, predicate);
            try {
                await promiseCancel.race(form.validateAll(names));
            }
            finally {
                form.submitting.value = false;
            }
            return resultFormData;
        },
        resetFields(formData) {
            if (form.submitting.value) {
                promiseCancel.cancelReject(new ValidationError());
            }
            if (formData === undefined) {
                form.resetFields();
            }
            else {
                resetFields(form, formData, transformedFormData);
            }
        },
        add(path$1, value) {
            const lastKey = path$1[path$1.length - 1];
            if (lastKey !== undefined) {
                const box = { [lastKey]: value };
                transformFormData(form, box);
                const transformedBox = box[lastKey];
                const valueAtPath = path(path$1, transformedFormData);
                if (Array.isArray(valueAtPath)) {
                    valueAtPath.push(transformedBox);
                }
                else {
                    set(transformedFormData, path$1, transformedBox);
                }
            }
        },
        remove(path$1) {
            const lastKey = path$1.pop();
            if (lastKey !== undefined) {
                if (path$1.length === 0) {
                    disposeForm(form, transformedFormData[lastKey]);
                    delete transformedFormData[lastKey];
                }
                else {
                    const valueAtPath = path(path$1, transformedFormData);
                    if (Array.isArray(valueAtPath)) {
                        const deletedFormData = valueAtPath.splice(+lastKey, 1);
                        disposeForm(form, deletedFormData);
                    }
                    else {
                        disposeForm(form, valueAtPath[lastKey]);
                        delete valueAtPath[lastKey];
                    }
                }
            }
        }
    };
}

/**
 *
 * Configure the validation behavior of `useValidation`.
 *
 * @param configuration - The Form validation configuration
 */
function createValidation(configuration) {
    return {
        install() {
            var _a;
            for (const [key, validationBehavior] of Object.entries((_a = configuration.validationBehavior) !== null && _a !== void 0 ? _a : {})) {
                VALIDATION_CONFIG.validationBehavior.set(key, validationBehavior);
            }
            if (VALIDATION_CONFIG.validationBehavior.has(configuration.defaultValidationBehavior)) {
                VALIDATION_CONFIG.defaultValidationBehavior =
                    configuration.defaultValidationBehavior;
            }
            else {
                console.warn(`[useValidation] Default validation behavior '${configuration.defaultValidationBehavior}' is not valid. Valid values are`, VALIDATION_CONFIG.validationBehavior.keys());
            }
        }
    };
}

exports.ValidationError = ValidationError;
exports.createValidation = createValidation;
exports.useValidation = useValidation;
